# Wand documentation build configuration file, created by
# sphinx-quickstart on Wed Sep 28 18:27:30 2011.
#
# This file is execfile()d with the current
# directory set to its containing directory.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))
sys.path.append(os.path.abspath('_themes'))
from wand.version import VERSION  # noqa

# Mocking C libraries to fake wand.api module which is unavailable
# on ReadTheDocs builder.
if os.environ.get('READTHEDOCS', 0):
    try:
        import wand.api  # noqa
    except ImportError:
        pass

    class Mock:
        def __init__(self, name):
            self.name = name

        def __getattr__(self, name):
            return Mock(self.name + '.' + name)

        def __repr__(self):
            return self.name

    mockapi = sys.modules['wand._api']
    mockapi.library = Mock('wand.api.library')
    mockapi.libmagick = Mock('wand.api.libmagick')
    mockapi.libc = Mock('wand.api.libc')
    sys.modules['wand'].api = sys.modules['wand.api'] = mockapi
    sys.modules['wand.version'].MAGICK_VERSION = None
    sys.modules['wand.version'].MAGICK_VERSION_NUMBER = 0x700  # Not break rtd
    sys.modules['wand.version'].MAGICK_VERSION_INFO = None
    sys.modules['wand.version'].MAGICK_RELEASE_DATE_STRING = None
    sys.modules['wand.version'].MAGICK_RELEASE_DATE = None
    sys.modules['wand.version'].QUANTUM_DEPTH = None
    sys.modules['wand.version'].QUANTUM_RANGE = None
    sys.modules['wand.version'].QUANTUM_SCALE = None
    sys.modules['wand.version'].MAGICK_HDRI = None

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx',
              'sphinx.ext.extlinks', 'sphinx.ext.todo']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Wand'
copyright = str(datetime.date.today().year) + ', Hong Minhee'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = VERSION
# The full version, including alpha/beta/rc tags.
release = VERSION

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''

# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`)
# to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'trac'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'flask'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {'index_logo': '', 'index_logo_height': ''}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['_themes']

html_style = 'wand.css'

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = project + ' ' + release

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = project

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/wand.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Wanddoc'


# -- Options for LaTeX output -------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
#  documentclass [howto/manual]).
latex_documents = [
    ('index', 'Wand.tex', 'Wand Documentation', 'Hong Minhee \\and E. McConville', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'wand', 'Wand Documentation',
     ['Hong Minhee', 'E. McConville'], 1)
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

extlinks = {
    'issue': ('https://github.com/emcconville/wand/issues/%s', '#%s'),
    'pull': ('https://github.com/emcconville/wand/pull/%s', '!%s'),
    'branch': ('https://github.com/emcconville/wand/compare/master...%s', '%s'),
    'commit': ('https://github.com/emcconville/wand/commit/%s', '%s'),
    'cli': ('https://imagemagick.org/script/command-line-options.php#%s', '-%s'),
}

# fall back if theme is not there
try:
    __import__('flask_theme_support')
except ImportError:
    print('-' * 74)
    print('Warning: Flask themes unavailable. Building with default theme')
    print('If you want the Flask themes, run this command and build again:')
    print('')
    print(' git submodule update --init')
    print('-' * 74)

    pygments_style = 'tango'
    html_theme = 'default'
    html_theme_options = {}
    html_logo = None
    html_style = None
